#include "MAX40080_PlatformSpecific.h"
#include "MAX40080.h"
#include "MAX31343_PlatformSpecific.h"
#include "hal_data.h"
#include "r_iic_master.h"

static MAX31343_Status MapStatus(MAX31343_Status status) {
	switch (status) {
		case MAX40080_Status_Ok: return MAX31343_Status_Ok;
		case MAX40080_Status_I2CNack: return MAX31343_Status_I2CNack;
		case MAX40080_Status_I2CError: return MAX31343_Status_I2CError;
		case MAX40080_Status_I2CTimeout: return MAX31343_Status_I2CTimeout;
		case MAX40080_Status_PacketErrorCheckFailed: return MAX31343_Status_I2CError;
		case MAX40080_Status_NotImplemented: return MAX31343_Status_NotImplemented;
		case MAX40080_Status_BadArg: return MAX31343_Status_BadArg;
		case MAX40080_Status_InvalidOperation: return MAX31343_Status_InvalidOperation;
		case MAX40080_Status_FifoIsEmpty: return MAX31343_Status_I2CError;
		case MAX40080_Status_NotSupported: return MAX31343_Status_InvalidOperation;
		default: return MAX31343_Status_NotImplemented;
	}
}

MAX31343_Status MAX31343_PlatformSpecific_Init() {
	MAX40080_Status status = MAX40080_PlatformSpecific_Init();
	return MapStatus(status);
}

MAX31343_Status MAX31343_PlatformSpecific_Deinit() {
	MAX40080_Status status = MAX40080_PlatformSpecific_Deinit();
	return MapStatus(status);
}

static i2c_master_event_t i2cEvent = 0;

static void MAX31343_PlatformSpecific_I2C_Callback(i2c_master_callback_args_t * p_args) {
	i2cEvent = p_args->event;
}

static MAX31343_Status MAX31343_PlatformSpecific_WaitForCompletion(i2c_master_event_t expectedEvent) {
	uint32_t timeout = 1000;
	while (i2cEvent == 0 && timeout--) {
		R_BSP_SoftwareDelay(1, BSP_DELAY_UNITS_MICROSECONDS);
		//__WFI();
	}

	if (i2cEvent == 0) {
		return MAX31343_Status_I2CTimeout;
	} else if (i2cEvent != expectedEvent) {
		R_IIC_MASTER_Abort(MAX31343_IIC_MASTER_INSTANCE.p_ctrl);
		return MAX31343_Status_I2CError;
	}

	return MAX31343_Status_Ok;
}

MAX31343_Status MAX31343_PlatformSpecific_ConfigureExternalInterrupt() {
	// this function should initialize hardware for 
	return MAX31343_Status_NotImplemented;
}

MAX31343_Status MAX31343_PlatformSpecific_ReadRegisters(uint8_t registerAddress, uint8_t* buffer, size_t bufferSize) {
	fsp_err_t fStatus;
	MAX31343_Status status;

	xSemaphoreTake(g_i2c_mutex, portMAX_DELAY);

	fStatus = R_IIC_MASTER_SlaveAddressSet(MAX31343_IIC_MASTER_INSTANCE.p_ctrl, MAX31343_I2C_7BIT_ADDRESS, I2C_MASTER_ADDR_MODE_7BIT);
	if (fStatus) {
		xSemaphoreGive(g_i2c_mutex);
		return MAX31343_Status_I2CError;
	}

	fStatus = R_IIC_MASTER_CallbackSet(MAX31343_IIC_MASTER_INSTANCE.p_ctrl, MAX31343_PlatformSpecific_I2C_Callback, NULL, NULL);
	if (fStatus) {
		xSemaphoreGive(g_i2c_mutex);
		return MAX31343_Status_I2CError;
	}

	i2cEvent = 0;
	fStatus = R_IIC_MASTER_Write(MAX31343_IIC_MASTER_INSTANCE.p_ctrl, &registerAddress, 1, 1);
	if (fStatus) {
		R_IIC_MASTER_Abort(MAX31343_IIC_MASTER_INSTANCE.p_ctrl);
		xSemaphoreGive(g_i2c_mutex);
		return MAX31343_Status_I2CError;
	}
	status = MAX31343_PlatformSpecific_WaitForCompletion(I2C_MASTER_EVENT_TX_COMPLETE);
	if (status) {
		xSemaphoreGive(g_i2c_mutex);
		return status;
	}

	i2cEvent = 0;
	fStatus = R_IIC_MASTER_Read(MAX31343_IIC_MASTER_INSTANCE.p_ctrl, buffer, bufferSize, 0);
	if (fStatus) {
		R_IIC_MASTER_Abort(MAX31343_IIC_MASTER_INSTANCE.p_ctrl);
		xSemaphoreGive(g_i2c_mutex);
		return MAX31343_Status_I2CError;
	}
	status = MAX31343_PlatformSpecific_WaitForCompletion(I2C_MASTER_EVENT_RX_COMPLETE);
	if (status) {
		xSemaphoreGive(g_i2c_mutex);
		return status;
	}

	xSemaphoreGive(g_i2c_mutex);
	return MAX31343_Status_Ok;
}

MAX31343_Status MAX31343_PlatformSpecific_WriteRegisters(uint8_t registerAddress, uint8_t* buffer, size_t bufferSize) {
	fsp_err_t fStatus;
	MAX31343_Status status;

	if (bufferSize > 64) {
		return MAX31343_Status_NotImplemented;
	}

	xSemaphoreTake(g_i2c_mutex, portMAX_DELAY);

	uint8_t bufferToTransmit[65];
	bufferToTransmit[0] = registerAddress;

	for (size_t i = 0; i < bufferSize; i++) {
		bufferToTransmit[i + 1] = buffer[i];
	}

	fStatus = R_IIC_MASTER_SlaveAddressSet(MAX31343_IIC_MASTER_INSTANCE.p_ctrl, MAX31343_I2C_7BIT_ADDRESS, I2C_MASTER_ADDR_MODE_7BIT);
	if (fStatus) {
		xSemaphoreGive(g_i2c_mutex);
		return MAX31343_Status_I2CError;
	}

	fStatus = R_IIC_MASTER_CallbackSet(MAX31343_IIC_MASTER_INSTANCE.p_ctrl, MAX31343_PlatformSpecific_I2C_Callback, NULL, NULL);
	if (fStatus) {
		xSemaphoreGive(g_i2c_mutex);
		return MAX31343_Status_I2CError;
	}

	i2cEvent = 0;
	fStatus = R_IIC_MASTER_Write(MAX31343_IIC_MASTER_INSTANCE.p_ctrl, bufferToTransmit, bufferSize + 1, 0);
	if (fStatus) {
		R_IIC_MASTER_Abort(MAX31343_IIC_MASTER_INSTANCE.p_ctrl);
		xSemaphoreGive(g_i2c_mutex);
		return MAX31343_Status_I2CError;
	}
	status = MAX31343_PlatformSpecific_WaitForCompletion(I2C_MASTER_EVENT_TX_COMPLETE);
	if (status) {
		xSemaphoreGive(g_i2c_mutex);
		return status;
	}

	xSemaphoreGive(g_i2c_mutex);
	return MAX31343_Status_Ok;
}
