#include "data_log_api.h"
#include "FreeRTOS.h"
#include "SEGGER_RTT/SEGGER_RTT.h"
#include "qspi_helper.h"
#include "r_qspi.h"
#include "MAX31343.h"
#include "common_data.h"
#include "data_log.h"
#include "ospi_ep.h"
#include "ospi_commands.h"

static int isTrickleChargerSet = 0;

#define DAYS_CAPACITY				31
#define DAY_LOG_OFFSET				((DayLog*)0x68000000)
#define DAY_LOG_SECTORS				2

static void wait_operation();
static DayLog* FindDayLog(uint8_t year, uint8_t month, uint8_t date);
static void EraseDayLog(DayLog* dl);

static Sample currentSample;

void DataLog_GetCurrentSample(Sample* out) {
	xSemaphoreTake(g_current_sample_mutex, portMAX_DELAY);
	memcpy(out, &currentSample, sizeof(Sample));
	xSemaphoreGive(g_current_sample_mutex);
}

int DataLog_GetDayLog(DayLog* daylog, uint8_t year, uint8_t month, uint8_t date) {
	int found = 1;
	xSemaphoreTake(g_daylog_flash_mutex, portMAX_DELAY);
	DayLog* dl = FindDayLog(year, month, date);
	if (dl == NULL) {
		found = 0;
	} else {
		memcpy(daylog, dl, sizeof(DayLog));
	}
	xSemaphoreGive(g_daylog_flash_mutex);
	return found;
}

void DataLog_ErasaAllDayLogs() {
	xSemaphoreTake(g_daylog_flash_mutex, portMAX_DELAY);

	for (int i = 0; i < DAYS_CAPACITY; i++) {
		DayLog* dl = (DayLog*)(((void*)(DAY_LOG_OFFSET)) + i * DEVICE_SECTOR_SIZE * DAY_LOG_SECTORS);
		SEGGER_RTT_printf(0, "Erasing daylog %p: %04d-%02d-%02d\r\n", dl, dl->year, dl->month, dl->date);
		EraseDayLog(dl);
	}

	xSemaphoreGive(g_daylog_flash_mutex);
}

static double currentTotalCurrent = 0;
static double currentTotalVoltage = 0;
static float currentMaxCurrent = 0;
static float currentMaxVoltage = 0;
static float currentMinCurrent = 0;
static float currentMinVoltage = 0;
static int currentSamplesCount = 0;

static uint8_t currentYear = 0xFF;
static uint8_t currentMonth = 0;
static uint8_t currentDate = 0;
static uint8_t currentHour = 0;
static uint8_t currentSubhour = 0;

static DayLog* FindDayLog(uint8_t year, uint8_t month, uint8_t date) {
	SEGGER_RTT_printf(0, "Searching for daylog: %04d-%02d-%02d.\r\n", (int)year + 2000, (int)month, (int)date);

	for (unsigned int i = 0; i < DAYS_CAPACITY; i++) {
		DayLog* dl = (DayLog*)(((uint8_t*)(DAY_LOG_OFFSET)) + i * DEVICE_SECTOR_SIZE * DAY_LOG_SECTORS);
		SEGGER_RTT_printf(0, "Daylog %p: %04d-%02d-%02d\r\n", dl, dl->year, dl->month, dl->date);
		if (dl->year == year && dl->month == month && dl->date == date) {
			SEGGER_RTT_printf(0, "Daylog found: %p.\r\n", dl);
			return dl;
		}
	}

	SEGGER_RTT_printf(0, "Daylog not found.\r\n");
	return NULL;
}

static void EraseDayLog(DayLog* dl) {
	fsp_err_t fStatus;
	uint8_t* mem = (uint8_t*)dl;

	SEGGER_RTT_printf(0, "Erasing daylog %p.\r\n", dl);

	for (int i = 0; i < DAY_LOG_SECTORS; i++) {
		fStatus = R_OSPI_Erase(&g_ospi_ctrl, mem, DEVICE_SECTOR_SIZE);
		if (fStatus) {
			SEGGER_RTT_printf(0, "Error while triggering erase operation. Status code: %d\r\n", fStatus);
		}
		wait_operation();

		mem += DEVICE_SECTOR_SIZE;
	}
}

static DayLog* RecycleDayLog() {
	SEGGER_RTT_printf(0, "Searching for daylog to recycle.\r\n");

	DayLog* min = DAY_LOG_OFFSET;
	if (min->year == 0xFF) {
		return min;
	}
	for (unsigned int i = 1; i < DAYS_CAPACITY; i++) {
		DayLog* dl = (DayLog*)(((uint8_t*)(DAY_LOG_OFFSET)) + i * DEVICE_SECTOR_SIZE * DAY_LOG_SECTORS);
		if (dl->year == 0xFF || dl->year < min->year) {
			min = dl;
		} else if (dl->year == min->year) {
			if (dl->month < min->month) {
				min = dl;
			} else if (dl->month == min->month) {
				if (dl->date < min->date) {
					min = dl;
				}
			}
		}
	}

	return min;
}

static void commitEntry() {
	fsp_err_t fStatus;

	xSemaphoreTake(g_daylog_flash_mutex, portMAX_DELAY);

	DayLog* dl = FindDayLog(currentYear, currentMonth, currentDate);
	if (dl == NULL) {
		dl = RecycleDayLog();
		SEGGER_RTT_printf(0, "Daylog for recycle: %p.\r\n", dl);

		EraseDayLog(dl);
	}

	uint8_t localMem[DAY_LOG_SECTORS * DEVICE_SECTOR_SIZE];
	DayLog* localDl = (DayLog*)localMem;
	memcpy(localDl, dl, DAY_LOG_SECTORS * DEVICE_SECTOR_SIZE);
	localDl->year = currentYear;
	localDl->month = currentMonth;
	localDl->date = currentDate;

	int entryIndex = currentHour * ENTRIES_PER_HOUR + currentSubhour;
	localDl->averageCurrent[entryIndex] = (float)(currentTotalCurrent / currentSamplesCount);
	localDl->averageVoltage[entryIndex] = (float)(currentTotalVoltage / currentSamplesCount);
	localDl->maximumCurrent[entryIndex] = currentMaxCurrent;
	localDl->maximumVoltage[entryIndex] = currentMaxVoltage;
	localDl->minimumCurrent[entryIndex] = currentMinCurrent;
	localDl->minimumVoltage[entryIndex] = currentMinVoltage;
	localDl->validData[entryIndex] = 1;

	SEGGER_RTT_printf(0, "Storing entry to daylog %p at position %d\r\n", dl, entryIndex);
	SEGGER_RTT_printf(0, "Storing AVG I=%.4f\r\n", localDl->averageCurrent[entryIndex]);
	SEGGER_RTT_printf(0, "Storing AVG U=%.4f\r\n", localDl->averageVoltage[entryIndex]);
	SEGGER_RTT_printf(0, "Storing MAX I=%.4f\r\n", localDl->maximumCurrent[entryIndex]);
	SEGGER_RTT_printf(0, "Storing MAX U=%.4f\r\n", localDl->maximumVoltage[entryIndex]);
	SEGGER_RTT_printf(0, "Storing MIN I=%.4f\r\n", localDl->minimumCurrent[entryIndex]);
	SEGGER_RTT_printf(0, "Storing MIN U=%.4f\r\n", localDl->minimumVoltage[entryIndex]);

	uint8_t* memDst = (uint8_t*)dl;
	uint8_t* memSrc = (uint8_t*)localMem;
	SEGGER_RTT_printf(0, "Struct size %d\r\n", (int)sizeof(DayLog));

	for (int i = 0; i < DAY_LOG_SECTORS; i++) {

		SEGGER_RTT_printf(0, "Erasing 4K region at %p\r\n", memDst);

		fStatus = R_OSPI_Erase(&g_ospi_ctrl, memDst, DEVICE_SECTOR_SIZE);
		if (fStatus) {
			SEGGER_RTT_printf(0, "Error while triggering erase operation. Status code: %d\r\n", fStatus);
		}
		wait_operation();

		for (unsigned int u = 0; u < DEVICE_SECTOR_SIZE / PAGE_WRITE_SIZE; u++) {

			SEGGER_RTT_printf(0, "Writing %d B region at src=%p dst=%p\r\n", PAGE_WRITE_SIZE, memSrc, memDst);

			fStatus = R_OSPI_Write(&g_ospi_ctrl, memSrc, memDst, PAGE_WRITE_SIZE);
			if (fStatus) {
				SEGGER_RTT_printf(0, "Error while writing data to FLASH. Status code: %d\r\n", fStatus);
			}
			wait_operation();

			memDst += PAGE_WRITE_SIZE;
			memSrc += PAGE_WRITE_SIZE;
		}
	}

	SEGGER_RTT_printf(0, "Veryfing %d B region at %p: ", DEVICE_SECTOR_SIZE, memDst - DEVICE_SECTOR_SIZE);
	if (memcmp(memSrc - DEVICE_SECTOR_SIZE, memDst - DEVICE_SECTOR_SIZE, DEVICE_SECTOR_SIZE) == 0) {
		SEGGER_RTT_printf(0, "OK\r\n");
	} else {
		SEGGER_RTT_printf(0, "Failure\r\n");
	}

	xSemaphoreGive(g_daylog_flash_mutex);
}

static void addEntry(Sample* sample, MAX31343_DateTime* timestamp) {

	uint8_t subhour = timestamp->minutes / 5;
	if (timestamp->year != currentYear || timestamp->month != currentMonth || timestamp->date != currentDate || timestamp->hours != currentHour || subhour != currentSubhour) {
		if (currentYear != 0xFF) {
			SEGGER_RTT_printf(0, "Commiting entry from %04d-%02d-%02d %d:shr_%02d to FLASH memory. Switching to entry %04d-%02d-%02d %d:shr_%02d.\r\n", currentYear, currentMonth, currentDate,
					currentHour, currentSubhour, timestamp->year, timestamp->month, timestamp->date, timestamp->hours, subhour);

			commitEntry();
		} else {
			SEGGER_RTT_printf(0, "Commiting entry from %04d-%02d-%02d %d:%02d to FLASH memory is skipped because it is initial entry.\r\n", currentYear, currentMonth, currentDate, currentHour,
					currentSubhour);

		}
		currentTotalCurrent = 0;
		currentTotalVoltage = 0;
		currentMaxCurrent = sample->current;
		currentMinCurrent = sample->current;
		currentMaxVoltage = sample->voltage;
		currentMinVoltage = sample->voltage;
		currentSamplesCount = 0;

		currentYear = timestamp->year;
		currentMonth = timestamp->month;
		currentDate = timestamp->date;
		currentHour = timestamp->hours;
		currentSubhour = subhour;
	}

	currentTotalCurrent += sample->current;
	currentTotalVoltage += sample->voltage;
	currentSamplesCount += 1;

	if (sample->current > currentMaxCurrent) {
		currentMaxCurrent = sample->current;
	}
	if (sample->current < currentMinCurrent) {
		currentMinCurrent = sample->current;
	}

	if (sample->voltage > currentMaxVoltage) {
		currentMaxVoltage = sample->voltage;
	}
	if (sample->voltage < currentMinVoltage) {
		currentMinVoltage = sample->voltage;
	}
}

// function inspired in Renesas example
static void wait_operation() {
	int timeout = 10000;
	spi_flash_status_t status;
	status.write_in_progress = true;

	do {
		R_OSPI_StatusGet(&g_ospi_ctrl, &status);
		if (status.write_in_progress) {
			vTaskDelay(pdMS_TO_TICKS(1));
		}
	} while (status.write_in_progress && timeout--);

	if (status.write_in_progress) {
		SEGGER_RTT_printf(0, "OSPI Operation timed out.\r\n");
	}
}

void data_log_entry(void* pvParameters) {
	BaseType_t status;
	MAX31343_Status mStatus;
	fsp_err_t fStatus;

	FSP_PARAMETER_NOT_USED(pvParameters);

	currentSample.current = 0;
	currentSample.voltage = 0;

	R_IOPORT_PinWrite(&g_ioport_ctrl, RESET_PIN, BSP_IO_LEVEL_LOW);
	R_BSP_SoftwareDelay(DELAY_TIME, BSP_DELAY_UNITS_MICROSECONDS);
	R_IOPORT_PinWrite(&g_ioport_ctrl, RESET_PIN, BSP_IO_LEVEL_HIGH);
	R_BSP_SoftwareDelay(DELAY_TIME, BSP_DELAY_UNITS_MICROSECONDS);

	fStatus = ospi_init();
	if (fStatus) {
		SEGGER_RTT_printf(0, "OSPI initialization failed. Status code: %d\r\n", fStatus);
	}

	fStatus = ospi_write_enable_and_verify(&g_ospi_ctrl, SPI_FLASH_PROTOCOL_EXTENDED_SPI);
	if (FSP_SUCCESS != fStatus) {
		SEGGER_RTT_printf(0, "OSPI ospi_write_enable_and_verify failed. Status code: %d\r\n", fStatus);
	}

	/* configure ospi in extended spi mode */
	fStatus = reconfigure_device();
	if (FSP_SUCCESS != fStatus) {
		SEGGER_RTT_printf(0, "OSPI reconfigure_device failed. Status code: %d\r\n", fStatus);
	}

	/*Read the device id*/
	uint32_t ospi_id = 0;
	fStatus = read_device_id(&ospi_id);
	if (FSP_SUCCESS != fStatus) {
		SEGGER_RTT_printf(0, "OSPI read_device_id failed. Status code: %d\r\n", fStatus);
	}

	/*Print device id and main menu*/
	SEGGER_RTT_printf(0, "OSPI ID: %u\r\n", ospi_id);

	while (1) {
		if (!isTrickleChargerSet) {
			mStatus = MAX31343_TricleCharger_SetModeAndEnable(MAX31343_TrickleChargerMode_3k_without_diode);
			if (mStatus) {
				SEGGER_RTT_printf(0, "MAX31343 Failed to configuring trickle charger settings. Status code: %d\r\n", mStatus);
				continue;
			}

			isTrickleChargerSet = 1;
		}

		Sample sample;
		status = xQueueReceive(g_new_samples_queue, (void*)&sample, 0);

		if (status == pdTRUE) {
			MAX31343_DateTime dt;

			// update currentSample variable
			xSemaphoreTake(g_current_sample_mutex, portMAX_DELAY);
			memcpy(&currentSample, &sample, sizeof(Sample));
			xSemaphoreGive(g_current_sample_mutex);

			mStatus = MAX31343_GetDateTime(&dt, 0);
			if (mStatus) {
				SEGGER_RTT_printf(0, "MAX31343 Failed while getting time. Status code: %d\r\n", mStatus);
			}

			// process new sample
			addEntry(&sample, &dt);
		}

		vTaskDelay(1);
	}
}
