#include <data_log_api.h>
#include "max40080_thread.h"
#include "MAX40080.h"
#include "SEGGER_RTT/SEGGER_RTT.h"
#include <stdio.h>

int isMax40080InterruptPending = 0;

void MAX40080_InterruptHandler(external_irq_callback_args_t * p_args) {
	isMax40080InterruptPending = 1;
}

static void HandleFifoWarning() {
	MAX40080_Status status;

	int isWritten = 0;

	for (int i = 0; i < 1; i++) {
		Sample sample;

		status = MAX40080_ReadCurrentAndVoltage(&sample.current, &sample.voltage);
		if (status) {
			SEGGER_RTT_printf(0, "Reading current and voltage failed. Error code: %d\r\n", status);
			continue;
		}

		char message[128];
		snprintf(message, sizeof(message), "MAX40080 DATA I=%.4fA, U=%.3fV\r\n", sample.current, sample.voltage);
		SEGGER_RTT_SetTerminal(1);
		SEGGER_RTT_WriteString(0, message);
		SEGGER_RTT_SetTerminal(0);

		xQueueSendToBack(g_new_samples_queue, &sample, 1);
	}

}

static void HandleFifoOvorflow() {
	// this can happen for example when debugger is enabled and program is
	// stopped for long period of time.
	MAX40080_FlushFifo();
}

static MAX40080_Status ReinitializeMAX40080() {
	MAX40080_Status status;

	// This function is called at two situations:
	// 1) at begining in time when sensor is configured first time
	// 2) when some unrecoverable error occur and for example I2C bus get stuck driven by someone.

	MAX40080_Deinit();

	// power cycle sensor by EN pin
	R_IOPORT_PinWrite(&g_ioport_ctrl, BSP_IO_PORT_02_PIN_05, 0);
	vTaskDelay(pdMS_TO_TICKS(1));
	R_IOPORT_PinWrite(&g_ioport_ctrl, BSP_IO_PORT_02_PIN_05, 1);
	vTaskDelay(pdMS_TO_TICKS(10));

	status = MAX40080_Init();
	if (status) {
		SEGGER_RTT_printf(0, "INIT Failed.\r\n");
		return status;
	}

	MAX40080_Configuration config;
	MAX40080_GetDefaultConfiguration(&config);
	config.operatingMode = MAX40080_OperationMode_SelectedActive_4_SPS;
	config.adcSampleRate = MAX40080_AdcSampleRate_Both_at_0_5_ksps;
	config.inputRange = MAX40080_InputRange_10mV;
	config.digitalFilter = MAX40080_DigitalFilter_NoAverage;

	status = MAX40080_SetConfiguration(&config);
	if (status) {
		SEGGER_RTT_printf(0, "Error while configuring MAX40080.\r\n");
		return status;
	}

	MAX40080_FifoConfiguration fifoConfig;
	MAX40080_GetFifoDefaultConfiguration(&fifoConfig);
	fifoConfig.storingMode = MAX40080_FifoStoringMode_CurrentAndVoltage;
	fifoConfig.overflowWarningLevel = 1;

	status = MAX40080_SetFifoConfiguration(&fifoConfig);
	if (status) {
		SEGGER_RTT_printf(0, "Error while configuring MAX40080 FIFO.\r\n");
		return status;
	}

	status = MAX40080_EnableInterrupts(MAX40080_Interrupt_FifoOverflowWarning | MAX40080_Interrupt_FifoOverflown);
	if (status) {
		SEGGER_RTT_printf(0, "Error while configuring MAX40080 Interrupts.\r\n");
		return status;
	}

	status = MAX40080_FlushFifo();
	if (status) {
		SEGGER_RTT_printf(0, "Error while flushing MAX40080 FIFO.\r\n");
		return status;
	}

	status = MAX40080_ClearPendingInterrupts(0xFF);
	if (status) {
		SEGGER_RTT_printf(0, "Error while clearing MAX40080 interrupt flags.\r\n");
		return status;
	}

	R_ICU_ExternalIrqEnable(&g_external_irq0_ctrl);

	return MAX40080_Status_Ok;
}

void max40080_thread_entry(void * pvParameters) {
	MAX40080_Status status;
	fsp_err_t fStatus;

	FSP_PARAMETER_NOT_USED(pvParameters);

	R_ICU_ExternalIrqOpen(&g_external_irq0_ctrl, &g_external_irq0_cfg);

	do {
		status = ReinitializeMAX40080();
		if (status) {
			SEGGER_RTT_printf(0, "MAX40080 Init Failed.\r\n");
			vTaskDelay(10);
		}
	} while (status);

	while (1) {
		bsp_io_level_t intLevel;

		fStatus = R_IOPORT_PinRead(&g_ioport_ctrl, BSP_IO_PORT_04_PIN_09, &intLevel);
		if (fStatus) {
			__BKPT(0);
		}

		if (isMax40080InterruptPending || intLevel == BSP_IO_LEVEL_HIGH) {

			//SEGGER_RTT_SetTerminal(1);
			//SEGGER_RTT_printf(0, "IP=%d, IL=%d\r\n", isMax40080InterruptPending)

			MAX40080_Interrupt ints;
			status = MAX40080_GetPendingInterrupts(&ints);
			if (status) {
				SEGGER_RTT_printf(0, "MAX40080 Error while getting pendiong interrupts.\r\n");
				do {
					status = ReinitializeMAX40080();
					if (status) {
						SEGGER_RTT_printf(0, "MAX40080 recovery after failure failed.\r\n");
					}
				} while (status);
				continue;
			}

			status = MAX40080_ClearPendingInterrupts(ints);
			if (status) {
				SEGGER_RTT_printf(0, "MAX40080 Error while clearing interrupts.\r\n");
				do {
					status = ReinitializeMAX40080();
					if (status) {
						SEGGER_RTT_printf(0, "MAX40080 recovery after failure failed.\r\n");
					}
				} while (status);
				continue;
			}

			isMax40080InterruptPending = 0;

			if (ints & MAX40080_Interrupt_FifoOverflown) {
				ints &= ~MAX40080_Interrupt_FifoOverflown;

				SEGGER_RTT_printf(0, "MAX40080 FIFO OVF\r\n");
				HandleFifoOvorflow();
			} else if (ints & MAX40080_Interrupt_FifoOverflowWarning) {
				ints &= ~MAX40080_Interrupt_FifoOverflowWarning;

				HandleFifoWarning();
			}

			if (ints) {
				SEGGER_RTT_printf(0, "MAX40080 Unprocesed INT.\r\n");
			}

		}

		vTaskDelay(1);
	}
}
