#include <data_log_api.h>
#include "qspi_helper.h"
#include "common_utils.h"
#include "hal_data.h"
#include "r_spi_flash_api.h"
#include "data_log.h"

static fsp_err_t QSPI_GetFlashStatus(void);
static fsp_err_t QSPI_SwitchToQSPIMode(void);

// funciton inspired in Renesas example
void QSPI_Init(void) {
	fsp_err_t err = FSP_SUCCESS;
	/*
	 *
	 uint8_t data_write[PAGE_WRITE_SIZE] = { RESET_VALUE, };
	 uint8_t verify_written_data[PAGE_WRITE_SIZE] = { RESET_VALUE, };
	 = STATUS_REG_PAYLOAD;
	 uint8_t * p_mem_addr = (uint8_t*) QSPI_DEVICE_START_ADDRESS;
	 */

	/*
	 * this needs to be done since QPI is set by user in configuration
	 * and it sets QPI only in MCU but not in flash device
	 * so as a system (MCU + QSPI flash device) QPI mode does not get set by
	 * simply calling only R_QSPI_Open in QPI mode.
	 * Rather QPI mode enabling has to be done in Flash device as well
	 * So opening the driver in extended SPI mode only
	 * and QPI mode is enabled when QSPI_SwitchToQSPIMode sub-function is called
	 */
	spi_flash_cfg_t l_qspi_cfg;
	memcpy((spi_flash_cfg_t*) &l_qspi_cfg, (spi_flash_cfg_t*) &g_qspi_cfg, sizeof(spi_flash_cfg_t));
	l_qspi_cfg.spi_protocol = SPI_FLASH_PROTOCOL_EXTENDED_SPI;

	R_QSPI_Close(&g_qspi_ctrl);

	/* open QSPI with local configuration  */
	err = R_QSPI_Open(&g_qspi_ctrl, &l_qspi_cfg);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_Open Failed\r\n");
		APP_ERR_TRAP(err);
	}

	SEGGER_RTT_printf(0, "Initializing QSPI Flash Memory\r\n");

	/* write enable for further operations */
	err = R_QSPI_DirectWrite(&g_qspi_ctrl, &(g_qspi_cfg.write_enable_command), 1, false);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_DirectWrite Failed\r\n");
		APP_ERR_TRAP(err);
	}
	err = QSPI_GetFlashStatus();
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("Failed to get status for QSPI operation\r\n");
		APP_ERR_TRAP(err);
	}

	/*
	 * write QSPI flash status register
	 * This is required to make sure the device is ready for general
	 * read write operation,
	 * This performs settings such as physical reset,WP hardware pin disable,
	 * block protection lock bits clearing.
	 * for more details please refer Mx25L data sheet.
	 */
	uint8_t data_sreg[3] = { 0x01, 0x40, 0x00 };
	err = R_QSPI_DirectWrite(&g_qspi_ctrl, data_sreg, sizeof(data_sreg), false);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_DirectWrite Failed\r\n");
		APP_ERR_TRAP(err);
	}
	err = QSPI_GetFlashStatus();
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("Failed to get status for QSPI operation\r\n");
		APP_ERR_TRAP(err);
	}

	/*
	 * Verifying data written to QSPI flash status register
	 * Step 1: - send command byte - 0x05
	 * through R_QSPI_DirectWrite with last argument set as true
	 * Step 2 - read data through R_QSPI_DirectRead
	 */
	uint8_t sreg_data = 0;
	err = R_QSPI_DirectWrite(&g_qspi_ctrl, &(g_qspi_cfg.status_command), 1, true);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_DirectWrite Failed\r\n");
		APP_ERR_TRAP(err);
	}

	/*
	 *  we should not call function QSPI_GetFlashStatus here
	 * because the CS line should not get interrupted between write read
	 *
	 * Also MCU <SFMCD register> is set as 0 when status register is read
	 * to resume in ROM access mode hence API direct read returns error as part
	 * of parameter check itself
	 */
	err = R_QSPI_DirectRead(&g_qspi_ctrl, &sreg_data, 1);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_DirectRead Failed\r\n");
		APP_ERR_TRAP(err);
	}
	err = QSPI_GetFlashStatus();
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("Failed to get status for QSPI operation\r\n");
		APP_ERR_TRAP(err);
	}

	/* verify read status register data */
	if (sreg_data != 0x40) {
		APP_ERR_PRINT("Failed to get value set in the status register \r\n");
		APP_ERR_TRAP(-1);
	}

	err = QSPI_SwitchToQSPIMode();
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("QSPI_SwitchToQSPIMode failed\r\n");
		APP_ERR_TRAP(err);
	}
}

// function coped from Renesas example
static fsp_err_t QSPI_GetFlashStatus(void) {
	spi_flash_status_t status = { .write_in_progress = true };
	int32_t time_out = (INT32_MAX);
	fsp_err_t err = FSP_SUCCESS;

	do {
		/* Get status from QSPI flash device */
		err = R_QSPI_StatusGet(&g_qspi_ctrl, &status);
		if (FSP_SUCCESS != err) {
			APP_ERR_PRINT("R_QSPI_StatusGet Failed\r\n");
			return err;
		}

		/* Decrement time out to avoid infinite loop in case of consistent failure */
		--time_out;

		if ( RESET_VALUE >= time_out) {
			APP_PRINT("\r\n ** Timeout : No result from QSPI flash status register ** \r\n");
			return FSP_ERR_TIMEOUT;
		}

	} while (false != status.write_in_progress);

	return err;
}

// function coped from Renesas example
static fsp_err_t QSPI_SwitchToQSPIMode(void) {
	fsp_err_t err = FSP_SUCCESS;
	uint8_t data_qpi_en = QSPI_MX25L_CMD_ENTER_QPI_MODE;

	err = R_QSPI_DirectWrite(&g_qspi_ctrl, &(g_qspi_cfg.write_enable_command), 1, false);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_DirectWrite Failed\r\n");
		return err;
	}
	err = QSPI_GetFlashStatus();
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("Failed to get status for QSPI operation\r\n");
		return err;
	}

	/* send QPI mode enable command in flash device
	 * Note - no status register read after this operation
	 * because flash device has gone in QPI mode
	 * and MCU at this point is in extended SPI mode only.
	 * vice versa same is applicable while exiting QPI mode too.
	 */
	err = R_QSPI_DirectWrite(&g_qspi_ctrl, &data_qpi_en, 1, false);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_DirectWrite Failed\r\n");
		return err;
	}

	/* Command byte transferred to flash-> NOW  set the QPI protocol in MCU run time */
	err = R_QSPI_SpiProtocolSet(&g_qspi_ctrl, SPI_FLASH_PROTOCOL_QPI);
	if (FSP_SUCCESS != err) {
		APP_ERR_PRINT("R_QSPI_SpiProtocolSet Failed\r\n");
	}

	return err;
}
